-- The addon namespace.
local addon, ns = ...

-- Define the manager class and put it in the addon namespace.
local Manager = {}
local ManagerMetatable = { __index = Manager }
ns.Manager = Manager

-- Constructor.
function Manager:new(shieldBarrierIcon, shieldBlockIcon)
	local self = {}
	setmetatable(self, ManagerMetatable)
	
	-- Initialize variables.
	self.Initialized = false
	self.Active = false
	self.PlayerName = UnitName("player") 
	self.CurrentRage = 0
	self.TotalBlocked = 0
	self.ShieldBlockActive = false
	self.ShieldBarrierActive = false
	self.BlockedAmounts = ns.LinkedList:new()
	self.ElapsedTime = 0.0
	self.ShieldBarrierIcon = shieldBarrierIcon
	self.ShieldBlockIcon = shieldBlockIcon
		
	-- Main frame (pun intended) for listening.
	self.Main =  CreateFrame("Frame", nil, UIParent)
	
	-- Event handler stuff. handlers are not fired when the addon is inactive, 
	-- listeners are. This is just my personal terminology.
	self.Handlers = {}
	self.Listeners = {}
	
	function self.Handlers.COMBAT_LOG_EVENT_UNFILTERED(sender, ...)
		if not self.ShieldBlockIcon.Active then
			return
		end
	
		local eventType = select(2, CombatLogGetCurrentEventInfo())
		local target = select(9, CombatLogGetCurrentEventInfo())
		local spellId = select(12, CombatLogGetCurrentEventInfo())
		
		if eventType == "SPELL_AURA_APPLIED" and spellId == self.ShieldBlockIcon.SpellId and target == self.PlayerName then
			self.ShieldBlockActive = true
			if not AuraUtil.FindAuraByName("Last Stand", "player") then
				self.TotalBlocked = 0
			end
		end 
		
		if eventType == "SPELL_AURA_APPLIED" and (spellId == 12975 and select(4, GetTalentInfo(4,3,1))) and target == self.PlayerName then
			self.ShieldBlockActive = true
			self.TotalBlocked = 0
		end 
		
		if eventType == "SPELL_AURA_REMOVED" and (spellId == self.ShieldBlockIcon.SpellId or (spellId == 12975 and select(4, GetTalentInfo(4,3,1)))) and target == self.PlayerName then
			if not AuraUtil.FindAuraByName("Shield Block", "player") and not AuraUtil.FindAuraByName("Last Stand", "player") then
				self.ShieldBlockActive = false
			elseif AuraUtil.FindAuraByName("Shield Block", "player") then
				self.TotalBlocked = 0
			end			
		end 
		
		if eventType == "SWING_DAMAGE" and target == self.PlayerName then
			local damage = select(12, CombatLogGetCurrentEventInfo()) or 0
			local blocked = select(16, CombatLogGetCurrentEventInfo()) or 0
			local absorbed = select(17, CombatLogGetCurrentEventInfo()) or 0

			self.BlockedAmounts:Add(damage + blocked + absorbed)
			
			if blocked and blocked > 0 and self.ShieldBlockActive then
				self.TotalBlocked = self.TotalBlocked + blocked
			end
		end
		
		if eventType == "SPELL_DAMAGE" and target == self.PlayerName then
			local damage = select(15, CombatLogGetCurrentEventInfo()) or 0
			local blocked = select(19, CombatLogGetCurrentEventInfo()) or 0
			local absorbed = select(20, CombatLogGetCurrentEventInfo()) or 0

			if (blocked and blocked > 0) or ShieldMaidSpellDatabase[spellId] then
				self.BlockedAmounts:Add(damage + blocked + absorbed)
			end
			
			if blocked and blocked > 0 then
				ShieldMaidSpellDatabase[spellId] = true
				
				if self.ShieldBlockActive then
					self.TotalBlocked = self.TotalBlocked + blocked
				end
			end
		end
		
		if eventType == "SWING_MISSED" and target == self.PlayerName then
			local missType = select(12, ...)
			local absorbed = select(15, ...) or 0
			if missType == "ABSORB" then
				self.BlockedAmounts:Add(absorbed)
			end
		end
		
		if eventType == "SPELL_MISSED" and target == self.PlayerName then
			local missType = select(15, ...)
			local absorbed = select(18, ...) or 0
			if missType == "ABSORB" and ShieldMaidSpellDatabase[spellId] then
				self.BlockedAmounts:Add(absorbed)
			end
		end
	end

	function self.Handlers.UNIT_POWER_UPDATE(sender, ...)
		local unitId = select(1, ...)
		if unitId == "player" then
			self.CurrentRage = UnitPower("player", SPELL_POWER_RAGE)
		end
	end

	function self.Handlers.PLAYER_REGEN_ENABLED(...)
		if ShieldMaidConfig.hiddenOutOfCombat then
			self.ShieldBarrierIcon:Hide()
			self.ShieldBlockIcon:Hide()
		end
	end

	function self.Handlers.PLAYER_REGEN_DISABLED(...)
		if ShieldMaidConfig.hiddenOutOfCombat then
			self.ShieldBarrierIcon:Show()
			self.ShieldBlockIcon:Show()
		end
	end

	function self.Listeners.ADDON_LOADED(sender, ...)
		local name = select(1, ...)
		if name == addon then
			self:LoadDefaultConfig()
			self:LoadSpellDatabase()
		end
	end
	
	function self.Listeners.PLAYER_LOGIN(...)
		self:Load()
	end
	
	function self.Listeners.UPDATE_SHAPESHIFT_FORM(...)
		self:Load()
	end
	
	function self.Listeners.ACTIVE_TALENT_GROUP_CHANGED(...)
		self:Load()
	end
	
	return self
end


function Manager:GetTalentMultiplier()
	--never surrender
	if (select(4, GetTalentInfo(4,2,1))) then
		return ((-1) * (UnitHealth("player"))/(UnitHealthMax("player")) + 2)
	else
	--indomitable or bolster
		return 1.00
	end
end


-- Calculates the max absorb value.
function Manager:CalculateMaxAbsorb()
	local baseAttackPower, positiveBuff, negativeBuff = UnitAttackPower("player")
	local attackPower = baseAttackPower + positiveBuff + negativeBuff
	local versatilityPercentage = GetCombatRatingBonus(CR_VERSATILITY_DAMAGE_DONE)
	local versatilityBonusPercentage = GetVersatilityBonus(CR_VERSATILITY_DAMAGE_DONE)
	local versatilityMultiplier = 1 + (versatilityPercentage + versatilityBonusPercentage) / 100.00

	if (UnitLevel("player") >= 111) then
		return (11 + attackPower * 3.5 * versatilityMultiplier * self:GetTalentMultiplier() * 1.14 * 1.3 )
	else
		return (11 + attackPower * 3.5 * versatilityMultiplier * self:GetTalentMultiplier() * 1.21 * 1.3 )
	end
end	

-- Calculates the estimated absorb value.
function Manager:CalculateEstimatedAbsorb()
	local baseAttackPower, positiveBuff, negativeBuff = UnitAttackPower("player")
	local attackPower = baseAttackPower + positiveBuff + negativeBuff
	local versatilityPercentage = GetCombatRatingBonus(CR_VERSATILITY_DAMAGE_DONE)
	local versatilityBonusPercentage = GetVersatilityBonus(CR_VERSATILITY_DAMAGE_DONE)
	local versatilityMultiplier = 1 + (versatilityPercentage + versatilityBonusPercentage) / 100.00
	if (UnitLevel("player") >= 111) then
		return (11 + attackPower * 3.5 * versatilityMultiplier * self:GetTalentMultiplier() * 1.14)
	else
		return (11 + attackPower * 3.5 * versatilityMultiplier * self:GetTalentMultiplier() * 1.21 )
	end
end	

-- Calculates the estimated block based on entries from the past 6 seconds in the linked list.
-- At the same time, entries more than 6 seconds old are removed.
function Manager:CalculateEstimatedBlock()
	local shieldBlock = GetShieldBlock()
	local shieldK = 1423
	local level = UnitLevel("player")
	
		if level == 120 then
			shieldK = 6300
		else
			shieldK = 1423
		end

	local blockDR = shieldBlock / (shieldBlock + shieldK)
	local critblockDR = blockDR * 2
	
	if (blockDR > 0.85) then
		blockDR = .85
	end
	if (critblockDR > 0.85) then
		critblockDR = .85
	end

	local sum = self.BlockedAmounts:Sum(6)
	local mastery, _ = GetMasteryEffect()
	local criticalBlockChance = mastery / 100
	local criticalBlock = sum * criticalBlockChance * critblockDR
	local normalBlock = sum * (1 - criticalBlockChance) * blockDR

	return criticalBlock + normalBlock
end

-- Tick function, called in updateInterval intervals.
function Manager:Tick()
	if self.ShieldBarrierIcon.Active and self.ShieldBlockIcon.Active then
		local estimatedAbsorb = self:CalculateEstimatedAbsorb()
		local estimatedBlock = self:CalculateEstimatedBlock()
		local estimatedMax = self:CalculateMaxAbsorb()
		self.ShieldBarrierIcon:Update(self.CurrentRage, estimatedAbsorb, estimatedBlock, estimatedMax)
		self.ShieldBlockIcon:Update(self.CurrentRage, self.TotalBlocked, estimatedAbsorb, estimatedBlock)
	end
end

-- Enables/disables the addon depending on class and spec. 
function Manager:Load()
	local class, _ = UnitClassBase("player")
	local spec = GetSpecialization()
	if class == "WARRIOR" and spec == 3 then
		self.Active = true
		
		if self.Active and not self.ShieldBarrierIcon.Active and not self.ShieldBlockIcon.Active then
			self.ShieldBarrierIcon:Reload()
			self.ShieldBlockIcon:Reload()
			self.ShieldBarrierIcon:Show()
			self.ShieldBlockIcon:Show()
			self.ShieldBarrierIcon.Active = true
			self.ShieldBlockIcon.Active = true
			
			-- In case we reload or change spec in the middle of combat, we check for it here.
			if UnitAffectingCombat("player") or not ShieldMaidConfig.hiddenOutOfCombat then
				self.ShieldBarrierIcon:Show()
				self.ShieldBlockIcon:Show()
				self.ShieldBlockIcon.Active = true
				self.ShieldBarrierIcon.Active = true
			else
				self.ShieldBarrierIcon:Hide()
				self.ShieldBlockIcon:Hide()
				self.ShieldBlockIcon.Active = false
				self.ShieldBarrierIcon.Active = false
			end
			
			ns.Console.Write("|c0000FF00Loaded|r")
			if (not self:GetMessage()) then
				C_Timer.After(1, function() ns.Console.Write("A new option to display the absorb value of Ignore Pain while Ignore Pain is active has been added. Recommended to set size to 60 or greater.")
				end)
				self:SetMessage(true)
			end
		end
		
		self:Tick()
		
	elseif self.Active then
		self.Active = false
		self.ShieldBarrierIcon:Hide()
		self.ShieldBlockIcon:Hide()
		self.ShieldBlockIcon.Active = false
		self.ShieldBarrierIcon.Active = false
		ns.Console.Write("|c00FF0000Unloaded|r")
	end
end

-- Handler that makes sure the Tick function is called every ns.Config.updateInterval seconds.
function Manager:OnUpdateHandler(sender, seconds)
	self.ElapsedTime = self.ElapsedTime + seconds
	while self.Active and self.ElapsedTime > ns.Config.updateInterval do
		self:Tick()
		self.ElapsedTime = self.ElapsedTime - ns.Config.updateInterval
	end
end

-- OnEvent handler that delegates to the correct handler. If inactive, we only handle ACTIVE_TALENT_GROUP_CHANGED.
function Manager:OnEventHandler(sender, event, ...)
	if self.Active and self.Handlers[event] then
		self.Handlers[event](sender, ...)
	end
	if self.Listeners[event] then
		self.Listeners[event](sender, ...)
	end
end

-- Registers update and event handlers.
function Manager:RegisterHandlers()
	for k, v in pairs(self.Handlers) do
		self.Main:RegisterEvent(k);
	end
	
	for k, v in pairs(self.Listeners) do
		self.Main:RegisterEvent(k);
	end

	self.Main:SetScript("OnUpdate", function(...) self:OnUpdateHandler(...) end)
	self.Main:SetScript("OnEvent", function(...) self:OnEventHandler(...) end)
end

-- Set the default configuration if no previous configuration has been saved.
function Manager:LoadSpellDatabase()
	if not ShieldMaidSpellDatabase then
		ShieldMaidSpellDatabase = {}
	end
end

-- Set the default configuration if no previous configuration has been saved.
function Manager:LoadDefaultConfig()
	if not ShieldMaidConfig then
		ShieldMaidConfig = {}
	end

	for k, v in pairs(ns.Config.Default) do
		if ShieldMaidConfig[k] == nil then
			ShieldMaidConfig[k] = v
		end
	end
end

-- Set the default configuration if no previous configuration has been saved.
function Manager:Reset()
	ShieldMaidConfig = {}
	for k, v in pairs(ns.Config.Default) do
		ShieldMaidConfig[k] = v
	end
	
	self.ShieldBarrierIcon:Reload()
	self.ShieldBlockIcon:Reload()
end

-- Locks the icons.
function Manager:Lock()
	self.ShieldBarrierIcon:Lock()
	self.ShieldBlockIcon:Lock()
end

-- Unlocks the icons for dragging.
function Manager:Unlock()
	self.ShieldBarrierIcon:Unlock()
	self.ShieldBlockIcon:Unlock()
end

-- Sets the icon sizes.
function Manager:SetSize(size)
	ShieldMaidConfig.size = size
	self.ShieldBarrierIcon:Reload()
	self.ShieldBlockIcon:Reload()
end

-- Sets the icon scale.
function Manager:SetScale(scale)
	ShieldMaidConfig.scale = scale
	self.ShieldBarrierIcon:Reload()
	self.ShieldBlockIcon:Reload()
end

-- Sets the icon margin.
function Manager:SetMargin(margin)
	ShieldMaidConfig.margin = margin
	self.ShieldBarrierIcon:Reload()
	self.ShieldBlockIcon:Reload()
end

-- Sets a value indicating whether to hide the icons when out of combat.
function Manager:SetHiddenOutOfCombat(value)
	ShieldMaidConfig.hiddenOutOfCombat = value
	if UnitAffectingCombat("player") or not ShieldMaidConfig.hiddenOutOfCombat then
		self.ShieldBarrierIcon:Show()
		self.ShieldBlockIcon:Show()
	else
		self.ShieldBarrierIcon:Hide()
		self.ShieldBlockIcon:Hide()
	end
end

function Manager:SetSecondaryBarrier(value)
	ShieldMaidConfig.secondaryBarrier = value
end

-- Sets a value indicating whether to show the icon frames.
function Manager:SetShowFrames(value)
	ShieldMaidConfig.showFrames = value
	self.ShieldBarrierIcon:Reload()
	self.ShieldBlockIcon:Reload()
end

-- Sets a value indicating whether to show the glow.
function Manager:SetShowGlow(value)
	ShieldMaidConfig.showGlow = value
	if not value then
		self.ShieldBarrierIcon:HideGlow()
		self.ShieldBlockIcon:HideGlow()
	end
end

function Manager:SetTruncatedNumbers(value)
	ShieldMaidConfig.truncatedNumbers = value
end

-- Sets value for a one time message
function Manager:SetMessage(value)
	ShieldMaidConfig.message = true
end

-- Gets the value for a one time message
function Manager:GetMessage()
	return ShieldMaidConfig.message
end

-- Sets a value indicating whether to show the cooldown "clock" on icons.
function Manager:SetShowCooldown(value)
	ShieldMaidConfig.showCooldown = value
end